#include <stdio.h>

#include "interface.h"

/*void eval_and_log(char* code) {*/
/*char* result = QTS_EvalToJSON(code);*/
/*fputs("-->\n", stdout);*/
/*fputs(result, stdout);*/
/*fputs("<--\n", stdout);*/
/*}*/

int main() {
  /*eval_and_log(*/
  /*"\"hello \" + \"world!\""*/
  /*);*/
  /*eval_and_log(*/
  /*"\"hello \" + \"worl"*/
  /*);*/

  return 0;
}
