.App {
  max-width: 756px;
  margin: 1em auto;
  padding: 1em;
}

textarea,
input,
pre {
  border: 1px solid #aaa;
  border-radius: 2px;
  padding: 0.75em;
  min-width: 100%;
  min-height: 5em;
  margin-top: 1em;
  box-shadow: 0px 0.25em 1.5em rgba(0, 0, 0, 0.07);
}
